from __future__ import annotations
import httpx
from .resource_abc import Resource
from typing import Dict, Any, List
from pydantic import BaseModel, Field
from enum import Enum
import copy

class CollectionType(str, Enum):
    Single = "Single"
    Array = "Array"

class LifeTime(str, Enum):
    Perpetual = "Perpetual"
    TimeVariant = "TimeVariant"

class FieldType(str, Enum):
    String = "String"
    Boolean = "Boolean"
    DateTime = "DateTime"
    Decimal =  "Decimal"

class FieldDefinition(BaseModel):
    name: str
    lifetime: LifeTime
    type: FieldType
    collectionType: CollectionType = CollectionType.Single
    required: bool
    description: str = ""

# These are optional in the API create and will be given default values. When read is called
# they will not be returned if they have the default value
DEFAULT_FIELD = {
    "collectionType": "Single",
    "description": ""
}

class EntityTypeResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    entityTypeName: str
    displayName: str
    description: str
    fieldSchema: List[FieldDefinition]

    def read(self, client, old_state) -> Dict[str, Any]:
        entitytype = old_state.entitytype
        return client.request("get", f"/api/api/customentities/entitytypes/{entitytype}").json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True)
        res = client.request("POST", "/api/api/customentities/entitytypes", json=desired).json()
        return {"entitytype": res["entityType"]}

    def update(self, client: httpx.Client, old_state):
        remote = self.read(client, old_state)
        # enrich remote fields with the default values if not present
        remote["fieldSchema"] = [rem | DEFAULT_FIELD for rem in remote["fieldSchema"]]
        desired = self.model_dump(mode="json", exclude_none=True)
        effective = remote | copy.deepcopy(desired)
        for i in range(0, len(self.fieldSchema)):
            if i < len(remote["fieldSchema"]):
                eff_field = remote["fieldSchema"][i] | desired["fieldSchema"][i]
                effective["fieldSchema"][i] = eff_field
        if effective == remote:
            return None
        res = client.request("PUT", f"/api/api/customentities/entitytypes/{old_state.entitytype}",
            json=desired).json()
        return {"entitytype": res["entityType"]}

    @staticmethod
    def delete(client, old_state):
        raise RuntimeError("Cannot delete a custom entity definition")

    def deps(self):
        return []



from types import SimpleNamespace
from httpx import Client as httpxClient
from abc import ABC, abstractmethod
from typing import Dict, Any, Union, Optional, Sequence


class Ref(ABC):
    id: str

    @abstractmethod
    def attach(self, client: httpxClient) -> None:
        pass

    def deps(self):
        return []


class Resource(ABC):
    id: str

    @abstractmethod
    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None|Dict[str,Any]:
        pass

    @abstractmethod
    def create(self, client: httpxClient) -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    def update(self, client: httpxClient, old_state) -> Union[None, Dict[str, Any]]:
        pass

    @staticmethod
    @abstractmethod
    def delete(client: httpxClient, old_state) -> None:
        pass

    @abstractmethod
    def deps(self) -> Sequence["Resource|Ref"]:
        pass



from __future__ import annotations
import httpx
import copy
from .resource_abc import Resource, Ref
from typing import Optional, Dict, List, Any, Sequence
from pydantic import BaseModel, field_serializer, model_validator, Field
from enum import Enum


class ActionId(BaseModel):
    """ActionId resource used with IdSelector

    Example
    -------
    >>> from fbnconfig.access import ActionId
    >>> ActionId(scope="myscope", activity="execute", entity="Feature")

    -------
    """
    scope: str
    activity: str
    entity: str


class MatchAllSelector(BaseModel):
    type_name: str = Field("matchAllSelectorDefinition", init=False, exclude=True)
    actions: List[ActionId]
    name: Optional[str] = None
    description: Optional[str] = None

class IdSelector(BaseModel):
    """IdSelector resource used with PolicyResource

    Example
    -------
    >>> from fbnconfig.access import IdSelector, ActionId
    >>> IdSelector(
            name="feature_id_selector",
            description="feature_id_selector",
            identifier={"scope": "myscope", "code": "mycode"},
            actions=[ActionId(scope="myscope", activity="execute", entity="Feature")])
    """
    type_name: str = Field("idSelectorDefinition", init=False, exclude=True)
    identifier: Dict[str, str]
    actions: List[ActionId]
    name: Optional[str] = None
    description: Optional[str] = None

class MetadataExpression(BaseModel):
    metadataKey: str
    operator: str
    textValue: str|None

class MetadataSelector(BaseModel):
    type_name: str = Field("metadataSelectorDefinition", init=False, exclude=True)
    actions: List[ActionId]
    name: str|None = None
    description: str|None = None
    expressions: List[MetadataExpression]

class PolicySelector(BaseModel):
    type_name: str = Field("policySelectorDefinition", init=False, exclude=True)
    actions: List[ActionId]
    name: str|None = None
    description: str|None = None
    identityRestriction: Dict[str, str]|None = None
    restrictionSelectors: Sequence[Selector]|None = None

    @field_serializer("restrictionSelectors", when_used="always")
    def serialize_selectors(self, selectors: Any):
        # convert array of selectors
        return [
            {selector.type_name: selector}
            for selector in selectors
        ]


Selector = IdSelector|MatchAllSelector|MetadataSelector|PolicySelector

class WhenSpec(BaseModel):
    """
    WhenSpec resource used with PolicyResource

    Example
    -------
    >>> from fbnconfig.access import WhenSpec
    >>>WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00")

    Notes
    -------
    When deactivate is not supplied, the policy is valid from time in activate until end of time
    """
    activate: str
    deactivate: Optional[str] = None


class Grant(str, Enum):
    """Type of grant used with PolicyResource

    Available values are: Allow, Deny and Undefined
    """
    ALLOW = "Allow"
    DENY = "Deny"
    UNDEFINED = "Undefined"


class PolicyRef(BaseModel, Ref):
    id: str = Field(exclude=True, init=True)
    code: str

    def attach(self, client):
        try:
            client.request("get", f"/access/api/policies/{self.code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Policy {self.code} not found")
            else:
                raise ex

# TODO: Needs scope added
class PolicyResource(BaseModel, Resource):
    """Manage a policy

    Attributes
    -------
    id: str
        Resource identifier
    """
    id: str = Field(exclude=True)
    code: str
    description: str
    applications: List[str]
    grant: Grant
    # dict of idselector is the old form for backward compatability
    # new form is array of Selector
    selectors: Sequence[Selector|Dict[str, IdSelector]]
    when: WhenSpec
    remote: Dict[str, Any] = Field(
        {}, exclude=True, init=False
    )  # will store the current state of the policy

    @field_serializer("selectors", when_used="always")
    def serialize_selectors(self, selectors: Any):
        # backward compatability when the user passes in their own
        # dict
        if all([isinstance(selector, dict) for selector in selectors]):
            return selectors
        # new version: convert list of selectors to list of dicts
        # where the key in each dict is the type of selector
        return [
            {selector.type_name: selector}
            for selector in selectors
        ]

    def read(self, client, old_state):
        existing_policy = client.request("get", f"/access/api/policies/{self.code}")
        self.remote = (
            existing_policy.json()
        )  # post is just `code` but get is `id: {scope:..., code:...}` and put has no code in the body
        self.remote.pop("id")
        self.remote.pop("links")

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True)
        client.request("POST", "/access/api/policies", json=desired)
        return {"id": self.id, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if old_state.code != self.code:
            raise (RuntimeError("Cannot change the code on a policy"))
        self.read(client, old_state)
        remote = copy.deepcopy(self.remote)
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"code"})

        if (
            desired["when"].get("deactivate", None) is None
        ):  # deactivate is defaulted on the server so not a difference unless we set it
            remote["when"].pop("deactivate")
        if desired == remote:
            return None
        client.request("put", f"/access/api/policies/{self.code}", json=desired)
        return {"id": self.id, "code": self.code}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/access/api/policies/{old_state.code}")

    def deps(self):
        return []


class PolicyCollectionRef(BaseModel, Ref):
    id: str = Field(exclude=True, init=True)
    code: str
    scope: str = "default"

    def attach(self, client):
        try:
            client.get(f"/access/api/policycollections/{self.code}", params={"scope": self.scope})
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"PolicyCollection {self.scope}/{self.code} not found")
            else:
                raise ex


class PolicyCollectionResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    # collections can be referenced by scope, but it can't be specified
    scope: str|None = Field(None, init=False, exclude=True)
    code: str
    description: str|None = None
    policies: Sequence[PolicyResource|PolicyRef] = []
    policyCollections: Sequence[PolicyCollectionResource|PolicyCollectionRef]|None = []
    metadata: Dict[str, List[str]]|None = None

    @field_serializer("policies", when_used="json")
    def serialize_policies(self, policies: Any):
        return [{"code": p.code, "scope": "default"} for p in policies]

    @field_serializer("policyCollections", when_used="json")
    def serialize_collections(self, collections: Any):
        return [{"code": p.code, "scope": "default"} for p in collections]

    def read(self, client, old_state) -> Dict[str, Any]:
        scope = old_state.scope
        code = old_state.code
        params = {"scope": scope} if scope is not None else None
        remote = client.get(f"/access/api/policycollections/{code}", params=params).json()
        remote.pop("links")
        remote.pop("id")
        return remote

    def create(self, client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True)
        res = client.request("POST", "/access/api/policycollections", json=desired)
        return res.json()["id"]

    def update(self, client, old_state) -> Dict[str, Any]|None:
        if old_state.code != self.code:
            self.delete(client, old_state)
            return self.create(client)
        remote = self.read(client, old_state)
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"code"})
        if desired == remote:
            return None
        client.request("put", f"/access/api/policycollections/{self.code}", json=desired)
        return {"code": self.code, "scope": old_state.scope}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/access/api/policycollections/{old_state.code}")

    def deps(self) -> List[Resource|Ref]:
        deps: List[Resource|Ref] = []
        for pol in self.policies:
            deps.append(pol)
        if self.policyCollections:
            for col in self.policyCollections:
                deps.append(col)
        return deps


class PolicyIdRoleResource(BaseModel):
    """Used to refer to a policy resource in a role resource
    """

    policies: Sequence[PolicyResource|PolicyRef]|None = []
    policyCollections: List[PolicyCollectionResource]|None = []

    @field_serializer("policies", when_used="json")
    def serialize_policy_identifiers(self, policies: List[PolicyResource|PolicyRef]):
        # this takes a policy resource as a dep, but it only needs to send the identifiers
        return [{"code": p.code, "scope": "default"} for p in policies]

    @field_serializer("policyCollections", when_used="json")
    def serialize_collection_identifiers(self, policies: Any):
        # this takes a policy resource as a dep, but it only needs to send the identifiers
        return [{"code": p.code, "scope": "default"} for p in policies]

class Permission(str, Enum):
    """Permission type used on a role resource
    """
    READ = "Read"
    WRITE = "Write"
    EXECUTE = "Execute"

class NonTransitiveSupervisorRoleResource(BaseModel):
    roles: List[RoleResource|RoleRef]

    @field_serializer("roles", when_used="always")
    def serialize_role_identifiers(self, roles: List[RoleResource|RoleRef]):
        return [{"code": r.code, "scope": "default"} for r in roles]

class RoleResourceRequest(BaseModel):
    nonTransitiveSupervisorRoleResource: NonTransitiveSupervisorRoleResource|None = None
    policyIdRoleResource: PolicyIdRoleResource|None = None


class RoleRef(BaseModel, Ref):
    """Reference an existing Role
    """
    id: str = Field(exclude=True, init=True)
    scope: str = "default"
    code: str

    def attach(self, client):
        try:
            params = {"scope": self.scope} if self.scope is not None else None
            client.request("get", f"/access/api/roles/{self.code}", params=params)
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Role {self.scope}/{self.code} not found")
            else:
                raise ex


class RoleResource(BaseModel, Resource):
    """Define a role resource
    """
    id: str = Field(exclude=True)
    code: str
    description: str|None = None
    policy_resource: PolicyIdRoleResource|None = None
    resource: None|RoleResourceRequest = None
    when: WhenSpec
    permission: Permission
    roleHierarchyIndex: Optional[int] = None
    remote: Dict[str, Any] = Field(None, exclude=True, init=False)

    @model_validator(mode="before")
    @classmethod
    def extract_policy_ids(cls, options: Dict[str, Any]) -> Dict[str, Any]:
        # if an old format policy_resource is passed, build the resource from it
        # if not use the resource directly
        if options.get("policy_resource", None) is None:
            return options
        options["resource"] = RoleResourceRequest(policyIdRoleResource=options["policy_resource"])
        return options


    def read(self, client, old_state):
        get = client.request("get", f"/access/api/roles/{self.code}")
        self.remote = get.json()
        self.remote.pop("id")
        self.remote.pop("links")

    def create(self, client):
        body = self.model_dump(mode="json", exclude={"policy_resource"}, exclude_none=True)
        client.request("POST", "/access/api/roles", json=body)
        return {"id": self.id, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if old_state.code != self.code:
            raise (RuntimeError("Cannot change the code on a role"))
        self.read(client, old_state)
        remote = copy.deepcopy(self.remote)
        desired = self.model_dump(mode="json", exclude={"code", "policy_resource"}, exclude_none=True)
        remote["when"].pop(
            "deactivate"
        )  # deactivate is defaulted on the server so not a difference unless we set it
        remote.pop("roleHierarchyIndex")  # set by the server
        if desired == remote:
            return None
        client.request("put", f"/access/api/roles/{self.code}", json=desired)
        return {"id": self.id, "code": self.code}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/access/api/roles/{old_state.code}")

    def deps(self):
        if self.resource is None or self.resource.policyIdRoleResource is None:
            return []
        res = self.resource.policyIdRoleResource
        pol_deps = [v for v in res.policies] if res.policies is not None else []
        col_deps = [v for v in res.policyCollections] if res.policyCollections is not None else []
        return pol_deps + col_deps



from __future__ import annotations
import httpx
from .resource_abc import Resource, Ref
from pydantic import BaseModel, Field, computed_field
from enum import Enum
from typing import Dict, Any, Union


class SetType(str, Enum):
    PERSONAL = "personal"
    SHARED = "shared"


class SetRef(BaseModel, Ref):
    """ Refer to an existing configuration set """
    id: str = Field(None, exclude=True, init=True)
    scope: str = Field(exclude=True, init=True)
    code: str = Field(exclude=True, init=True)
    type: SetType

    def attach(self, client):
        # just check it exists
        scope = self.scope
        code = self.code
        type = self.type.value
        try:
            client.get(f"/configuration/api/sets/{type}/{scope}/{code}").json()
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Config set {type}/{scope}/{code} not found")
            else:
                raise ex


class SetResource(BaseModel, Resource):
    """ Manage a configuration set """
    id: str = Field(None, exclude=True, init=True)
    scope: str = Field(exclude=True, init=True)
    code: str = Field(exclude=True, init=True)
    description: str
    type: SetType
    _remote: Dict[str, Any] | None = None

    @computed_field(alias="id")
    def setId(self) -> Dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self, client, old_state):
        scope = self.scope
        code = self.code
        type = old_state.type
        get = client.get(f"/configuration/api/sets/{type}/{scope}/{code}")
        self._remote = get.json()

    def create(self, client: httpx.Client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.request("post", "/configuration/api/sets", json=desired)
        return {"scope": self.scope, "code": self.code, "type": self.type.value}

    @staticmethod
    def delete(client, old_state):
        type = old_state.type
        scope = old_state.scope
        code = old_state.code
        client.request("delete", f"/configuration/api/sets/{type}/{scope}/{code}")

    def update(self, client, old_state) -> Dict[str, Any] | None:
        if [old_state.scope, old_state.code, old_state.type] != [self.scope, self.code, self.type.value]:
            raise RuntimeError("Cannot change the scope, code or type on a config set")
        self.read(client, old_state)
        assert self._remote is not None
        current = {"description": self._remote["description"]}
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        desired.pop("id")
        desired.pop("type")
        if (desired == current):
            return None
        client.put(f"/configuration/api/sets/{self.type.value}/{self.scope}/{self.code}", json=desired)
        return {"scope": self.scope, "code": self.code, "type": self.type.value}

    def deps(self):
        return []


class ItemRef(BaseModel, Ref):
    """ Reference an existing configuration item with a set """
    id: str = Field(None, exclude=True, init=True)
    set: SetRef | SetResource
    key: str
    ref: str = Field(None, exclude=False, init=False)

    def attach(self, client):
        type = self.set.type.value
        scope = self.set.scope
        code = self.set.code
        key = self.key
        try:
            get = client.get(f"/configuration/api/sets/{type}/{scope}/{code}/items/{key}")
            self.ref = get.json()["ref"]
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError("Config item not found")
            else:
                raise ex


class ValueType(str, Enum):
    TEXT = "text"


class ItemResource(BaseModel, Resource):
    """ Manage a configuration item with a set """
    id: str = Field(None, exclude=True, init=True)
    set: SetRef | SetResource = Field(exclude=True, init=True)
    key: str
    ref: str = Field(None, exclude=False, init=False)
    value: Any
    valueType: ValueType
    isSecret: bool
    description: str
    _remote: Dict[str, Any] = {}

    def read(self, client, old_state):
        type = old_state.type
        scope = old_state.scope
        code = old_state.code
        key = old_state.key
        get = client.get(f"/configuration/api/sets/{type}/{scope}/{code}/items/{key}")
        self._remote = get.json()

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True)
        type = self.set.type.value
        scope = self.set.scope
        code = self.set.code
        post = client.post(
            f"/configuration/api/sets/{type}/{scope}/{code}/items",
            json=desired
        )
        match = next((item for item in post.json()["items"] if item["key"] == self.key), None)
        if match is None:
            raise RuntimeError("Something wrong creating config item")
        self.ref = match["ref"]
        return {"scope": scope, "code": code, "type": type, "key": self.key}

    @staticmethod
    def delete(client, old_state):
        type = old_state.type
        scope = old_state.scope
        code = old_state.code
        key = old_state.key
        client.delete(f"/configuration/api/sets/{type}/{scope}/{code}/items/{key}")

    def update(self, client, old_state):
        type = self.set.type.value
        scope = self.set.scope
        code = self.set.code
        key = self.key
        # if the location has changed we remove the old one
        if [old_state.scope, old_state.code, old_state.type] != [scope, code, type]:
            self.delete(client, old_state)
            return self.create(client)
        self.read(client, old_state)
        # can't update these using PUT
        if self.isSecret != self._remote["isSecret"] or self.valueType != self._remote["valueType"]:
            self.delete(client, old_state)
            return self.create(client)
        self.ref = self._remote["ref"]
        current = {k: v for k, v in self._remote.items() if k in ["description", "value"]}
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        desired = {k: v for k, v in desired.items() if k in ["description", "value"]}
        if desired == current:
            return None
        client.put(f"/configuration/api/sets/{type}/{scope}/{code}/items/{key}", json=desired)
        return {"scope": scope, "code": code, "type": type, "key": self.key}

    def deps(self):
        return [self.set]


class SystemConfigResource(BaseModel, Resource):
    """ Manage a system configuration item

        The default value is used to reset the system configuration value when the resource is deleted

       Example
       -------
           >>> from fbnconfig import Deployment
           >>> from fbnconfig.configuration import SystemConfigResource
           >>> validate_instruments = SystemConfigResource(
                    id="validate-instr",
                    code="TransactionBooking",
                    key="ValidateInstruments",
                    value=True,
                    description="Test from fbnconfig",
                    default_value=False)
           >>> Deployment("myDeployment", [validate_instruments])

       Attributes
       ----------
       id : str
         Resource identifier; this will be used in the log to reference the item resource
       code : str
         Code of the system configuration set; System configurations exist in the 'system' scope
       key: str
           Key of the set to use
       value: Any
           Configuration item value
       default_value: Any
         The value this configuration item will be set to when the resource is deleted
         """
    id: str = Field(None, exclude=True, init=True)
    code: str = Field(None, exclude=True, init=True)
    key: str
    value: str
    default_value: Any = Field(None, exclude=True, init=True)
    description: str
    _remote: Dict[str, Any] = {}

    def read(self, client, old_state):
        code = old_state.code
        key = old_state.key
        get = client.get(f"/configuration/api/sets/system/{code}/items/{key}")
        self._remote = get.json()["values"][0]

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True)
        result = client.put(
            f"/configuration/api/sets/shared/system/{self.code}/items/{self.key}",
            json=desired
        )
        if result is None:
            raise RuntimeError("Something wrong creating config item")
        return {"code": self.code, "key": self.key, "default_value": self.default_value}

    @staticmethod
    def delete(client, old_state):
        if old_state.default_value is None:
            pass
        else:
            desired = {"value": old_state.default_value}
            client.put(
                f"/configuration/api/sets/shared/system/{old_state.code}/items/{old_state.key}",
                json=desired)

    def update(self, client, old_state) -> Union[None, Dict[str, Any]]:
        code = self.code
        key = self.key

        if self.key != old_state.key or code != old_state.code:
            self.delete(client, old_state)
            return self.create(client)

        self.read(client, old_state)

        current = {k: v for k, v in self._remote.items() if k in ["description", "value"]}
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        desired = {k: v for k, v in desired.items() if k in ["description", "value"]}
        if desired == current:
            return None
        client.put(f"/configuration/api/sets/shared/system/{code}/items/{key}", json=desired)
        return {"code": code, "key": self.key, "default_value": self.default_value}

    def deps(self):
        return []



import time
import textwrap
import json
import sys
from typing import Dict, Any, List
from pydantic import BaseModel, Field
from .resource_abc import Resource, Ref
from enum import Enum

def background_query(client, sql):
    res = client.put("/honeycomb/api/SqlBackground", content=sql, headers={"Content-type": "text/plain"})
    return res.json()

def get_progress(client, execution_id, progress_url):
    return client.get("/honeycomb" + progress_url).json()

def wait_for_background(client, execution):
    execution_id = execution["executionId"]
    progress_url = execution["progress"]["href"]
    progress = get_progress(client, execution_id, progress_url)
    status = progress["status"]
    match status:
      case "Faulted" | "Cancelled":
          raise RuntimeError("Query was " + status + progress["progress"])
      case "Created" | "WaitingForActivation" | "WaitingToRun" | "WaitingForChildrenToComplete":
          return None
      case "Running":
          return None
      case "RanToCompletion":
          return progress
      case _:
          raise RuntimeError("Unknown status:" + status)

def fetch(client, execution, progress):
    res = client.get("/honeycomb" + execution["fetchJsonProper"]["href"])
    return res.json()

def query(client, sql):
    execution = background_query(client, sql)
    progress = None
    while True:
        if progress is None:
            time.sleep(1)
            progress = wait_for_background(client, execution)
        else:
            return fetch(client, execution, progress)

def put_query(client, sql):
    res = client.put(
        "/honeycomb/api/Sql/json",
        content=sql,
        params={"jsonProper": True},
        headers={"Content-type": "text/plain"}
    )
    return res.json()

class ParameterType(str, Enum):
    BigInt   =  "BigInt"
    Boolean  =  "Boolean"
    Date     =  "Date"
    DateTime =  "DateTime"
    Decimal  =  "Decimal"
    Double   =  "Double"
    Int      =  "Int"
    Table    =  "Table"
    Text     =  "Text"

class VariableType(str, Enum):
    Scalar = "@@"
    Table = "@"


class Variable(BaseModel):
    name: str
    type: VariableType
    sql: str

    def init_str(self):
        #  eg @scalar = select 2 + 2
        return f"{self.type.value}{self.name} = {self.sql}"

    def with_str(self):
        return f"{self.type.value}{self.name}"

class Parameter(BaseModel):
    name: str
    type: ParameterType
    value: Any
    setAsDefaultValue: bool = True
    isMandatory: bool = True
    tooltip: str

    # return in the same format thst sys.file stores it in
    def metadata(self):
        base: Dict[str, Any] = {
            "Name": self.name,
            "Type": self.type.value,
            "Description": self.tooltip,
        }
        if self.setAsDefaultValue and self.type != ParameterType.Table:
            base["DefaultValue"] = self.value
        if self.type == ParameterType.Table:
            if self.isMandatory:
                base["ConditionUsage"] = 2
            else:
                base["ConditionUsage"] = 0
        return base

def lumi_fmt(value):
    if isinstance(value, Variable):
        return f"{value.type.value}{value.name}"
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float)):
        return str(value)
    return '"' + value.replace('"', '""') + '"'


class ViewRef(BaseModel, Ref):
    """ Reference an existing view


    Example
    ----------
    >>> from fbnconfig import lumi
    >>> lumi.ViewResource(
    ...  id="lumi-example-ref",
    ...  provider="Views.fbnconfig.existing_view")


    Attributes
    ----------
    id : str
         Resource identifier.
    provider : str
        Name of the view referenced. This is assumed to exist
    """
    id: str = Field(exclude=True)
    provider: str

    def attach(self, client):
        path = self.provider.replace(".", "/")
        res = put_query(
            client,
            f"select 1 from sys.file where path = 'databaseproviders/{path}.sql'"
        )
        if len(res) != 1:
            raise RuntimeError(f"Failed to attach ref to {self.provider}. No rows returned")


class ViewResource(BaseModel, Resource):
    """Create and manage a Luminesce view

    Example
    ----------
    >>> from fbnconfig import lumi
    >>> lumi.ViewResource(
    ...  id="lumi-example-view",
    ...  provider="Views.fbnconfig.example",
    ...  description="My resource test view",
    ...  documentationLink="http://example.com/query",
    ...  variableShape=False,
    ...  useDryRun=True,
    ...  allowExecuteIndirectly=False,
    ...  distinct=True,
    ...  sql='select 2+#PARAMETERVALUE(p1)   as   twelve',
    ...  parameters=[
    ...      lumi.Parameter(
    ...          name="p1",
    ...          value=10,
    ...          setAsDefaultValue=True,
    ...          tooltip="a number",
    ...          type=lumi.ParameterType.Int
    ...      )

    Attributes
    ----------
    id : str
         Resource identifier.
    provider : str
        Name of the view managed by this resource
    description : str
        View description
    sql: str
        The query string for the view
    parameters : list of `Parameter`, optional
        List of parameters for the view
    dependencies : list of dependencies, optional
        This can be another view or any other resource
    documentationLink: str, optional
        Displays one or more hyperlinks in the summary dialog for the view
    variableShape: bool, optional
        This is useful if data returned is likely to vary in shape between queries. Defaults to false.
    allowExecuteIndirectly : bool, optional
        Allows end users to query providers within the view even if they are not entitled to use those
        providers directly.
        Defaults to false.
    limit: int, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    groupBy: str, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    filter: str, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    offset: int, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    distinct: bool, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    useDryRun: bool, optional
        Intended for automatic deployment of views. See docs for more details. Defaults to false
    variables: List of `Variable`, optional
        A table variable that can be passed into the view by an end user or in code

    See Also
    --------
    `https://support.lusid.com/knowledgebase/article/KA-01767/en-us`__
    """
    id: str = Field(exclude=True)
    provider: str = Field(serialization_alias="Provider")
    description: str = Field(serialization_alias="Description")
    sql: str
    parameters: List[Parameter] = []
    dependencies: List | None = None
    documentationLink: str|None = None
    variableShape: bool|None = None
    allowExecuteIndirectly: bool|None = None
    limit: int|None = None
    groupBy: str|None = None
    filter: str|None = None
    offset: int|None = None
    distinct: bool|None = None
    useDryRun: bool|None = None
    variables: List[Variable] = []

    class Registration:
        tries = 10
        wait_time = 1

    _saved_options = { # maps from sys.file metadata to view option names
        "Description": "description",
        "DocumentationLink": "documentationLink",
        "IsWithinDirectProviderView": "variableShape",
        "IsWithinViewAllowingIndirectExecute": "allowExecuteIndirectly",
    }
    _test_options = [
        "distinct",
        "filter",
        "groupby",
        "limit",
        "offset",
        "preamble",
        "useDryRun",
    ]

    def read(self, client, old_state) -> Dict[str, Any]:
        path = old_state.provider.replace(".", "/")
        res = put_query(
            client,
            textwrap.dedent(f"""\
                select f.Content, r.Version from sys.file as f
                join sys.registration as r on r.Name = '{old_state.provider}'
                where path = 'databaseproviders/{path}.sql'
                order by r.Version asc
                limit 1
            """)
        )
        assert len(res) == 1

        def strip_column_description(kv: Dict) -> Dict:
            if kv["Type"] == "Table":
                kv["Description"] = kv["Description"].split("\nAvailable columns")[0]

            return kv

        parts = res[0]["Content"].split("--- MetaData ---")
        sql = parts[0]
        metadata = json.loads(parts[1])

        parameters =  [strip_column_description(p) for p in metadata["Parameters"]]
        props = {
            v: metadata[k]
            for k, v in self._saved_options.items()
            if metadata.get(k, None) is not None
        }
        return {"sql": sql, "version": res[0]["Version"], "parameters": parameters} | props

    @staticmethod
    def registration_version(client, view_name) -> int | None:
        rows = put_query(client, textwrap.dedent(f"""\
            select Version from sys.registration where Name='{view_name}'
            order by Version asc
            limit 1
        """))
        return int(rows[0]["Version"]) if len(rows) > 0 else None

    def format_option(self, option, value):
        if isinstance(value, bool) and value:
            return f"--{option}"
        if isinstance(value, (int, float)):
            return f"--{option}={value}"
        return f"--{option}={lumi_fmt(value)}"

    def get_variables(self):
        param_variables = [
            param.value for param in self.parameters
            if isinstance(param.value, Variable)
        ]
        seen = set()
        return [
            value for value in self.variables + param_variables
            if value.name not in seen and not seen.add(value.name)
        ]

    def template(self, desired):
        options = [
            self.format_option(option, desired[option])
            for option in ["provider"] + self._test_options + list(self._saved_options.values())
            if desired.get(option) is not None
        ]
        tpl = textwrap.dedent("""\
            {preamble}@x = use Sys.Admin.SetupView{with_clause}
            {options}{params}
            ----
            {sql}
            enduse;
            select * from @x;
        """)
        params = [
            # can't quote the tooltip as causes an error from lumi HC-3322
            f"{p.name},{p.type.value},{lumi_fmt(p.value)},{lumi_fmt(p.isMandatory)},{p.tooltip}"
            if p.type == ParameterType.Table else
            f"{p.name},{p.type.value},{lumi_fmt(p.value)},{lumi_fmt(p.setAsDefaultValue)},{p.tooltip}"
            for p in self.parameters
        ]
        param_clause = "\n--parameters\n{0}".format("\n".join(params)) \
            if len(params) > 0 else ""
        variables = self.get_variables()
        preamble = ";\n".join([v.init_str() for v in variables]) + ";\n" \
            if len(variables) > 0 else ""
        with_clause = " with " + ", ".join([v.with_str() for v in variables]) \
            if len(variables) > 0 else ""
        sql = tpl.format(
            options="\n".join(options),
            params=param_clause,
            sql=desired["sql"],
            with_clause=with_clause,
            preamble=preamble
        )
        return sql

    def create(self, client) -> Dict[str, Any]:
        desired = self.model_dump(exclude_none=True)
        sql = self.template(desired)
        put_query(client, sql)
        for i in range(0, self.Registration.tries):
            if self.registration_version(client, self.provider) is not None:
                break
            else:
                if i == self.Registration.tries - 1:
                    sys.stderr.write(
                        f"warning: no view registration after {i} tries for {self.provider}"
                    )
                else:
                    time.sleep(self.Registration.wait_time)
        return {"provider": self.provider}

    def update(self, client, old_state):
        if self.provider != old_state.provider:
            self.delete(client, old_state)
            self.create(client)
            return
        desired = self.model_dump(exclude_none=True, by_alias=False, exclude=set(self._test_options))
        raw_remote = self.read(client, old_state)
        remote_props = {
            k.lower(): v
            for k, v in raw_remote.items()
            if k in self._saved_options.values() and v is not None
        }
        desired_props = {
            k.lower(): v
            for k, v in desired.items()
            if k in self._saved_options.values() and v is not None
        }
        remote_params = raw_remote["parameters"]
        desired_params = [p.metadata() for p in self.parameters]
        effective_params = [
            (remote_params[i] | desired_params[i])
            if i < len(remote_params) and remote_params[i]["Name"] == desired_params[i]["Name"]
            else desired_params[i]
            for i, des in enumerate(desired_params)
        ]
        remote_sql = textwrap.dedent(raw_remote["sql"].rstrip())
        remote_version = raw_remote["version"]
        desired_sql = textwrap.dedent(self.sql.rstrip())
        if (desired_sql == remote_sql and remote_props | desired_props == remote_props
                and effective_params == remote_params):
            return None
        sql = self.template(desired)
        put_query(client, sql)
        for i in range(0, self.Registration.tries):
            version: int | None = self.registration_version(client, self.provider)
            if version is not None and remote_version < version:
                break
            else:
                if i == self.Registration.tries - 1:
                    sys.stderr.write(
                        f"warning: no view registration after {i} tries for {self.provider}"
                    )
                else:
                    time.sleep(self.Registration.wait_time)
        return {"provider": self.provider}

    def deps(self):
        return self.dependencies if self.dependencies else []

    @staticmethod
    def delete(client, old_state):
        sql = textwrap.dedent(f"""\
        @x = use Sys.Admin.SetupView
        --provider={old_state.provider}
        --deleteProvider
        ----
        select 1 as deleting
        enduse;
        select * from @x;
        """)
        put_query(client, textwrap.dedent(sql))
        for i in range(0, ViewResource.Registration.tries):
            if ViewResource.registration_version(client, old_state.provider) is None:
                break
            else:
                if i == ViewResource.Registration.tries - 1:
                    sys.stderr.write(
                        f"warning: no view deregistration after {i} tries for {old_state.provider}"
                    )
                else:
                    time.sleep(ViewResource.Registration.wait_time)



from __future__ import annotations
import httpx
from .resource_abc import Resource, Ref
import typing
from typing import Dict, Any
from pydantic import BaseModel, Field
from enum import Enum
import copy

class TypeValueRange(str, Enum):
    Open = "Open"
    Closed = "Closed"


class ValueType(str, Enum):
    Boolean              = "Boolean"
    Code                 = "Code"
    Currency             = "Currency"
    CurrencyAndAmount    = "CurrencyAndAmount"
    CutLocalTime         = "CutLocalTime"
    DateOrCutLabel       = "DateOrCutLabel"
    DateTime             = "DateTime"
    Decimal              = "Decimal"
    Id                   = "Id"
    Int                  = "Int"
    List                 = "List"
    Map                  = "Map"
    MetricValue          = "MetricValue"
    Percentage           = "Percentage"
    PropertyArray        = "PropertyArray"
    ResourceId           = "ResourceId"
    ResultValue          = "ResultValue"
    String               = "String"
    TradePrice           = "TradePrice"
    UnindexedText        = "UnindexedText"
    Uri                  = "Uri"

class Unit(BaseModel):
    code: str
    displayName: str
    description: str
    details: Any|None = None

class UnitSchema(str, Enum):
    NoUnits = "NoUnits"
    Basic = "Basic"
    Iso4217Currency = "Iso4217Currency"

class FieldDefinition(BaseModel):
    key: str
    isRequired: bool
    isUnique: bool

class FieldValue(BaseModel):
    value: str
    fields: Dict[str, str]  # limit to strings because get returns strings

class ReferenceData(BaseModel):
    fieldDefinitions: typing.List[FieldDefinition]
    values: typing.List[FieldValue]

class DataTypeRef(BaseModel, Ref):
    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        scope, code = self.scope, self.code
        try:
            client.get(f"/api/api/datatypes/{scope}/{code}", params={"includeSystem": True})
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Datatype {scope}/{code} not found")
            else:
                raise ex

class DataTypeResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str
    code: str
    typeValueRange: TypeValueRange
    displayName: str
    description: str
    valueType: ValueType
    acceptableValues: typing.List[str]|None = None
    unitSchema: UnitSchema = UnitSchema.NoUnits
    acceptableUnits: typing.List[Unit]|None = None
    referenceData: ReferenceData|None = None

    def read(self, client, old_state) -> Dict[str, Any]:
        return client.request("get", f"/api/api/datatypes/{old_state.scope}/{old_state.code}").json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True)
        client.request("POST", "/api/api/datatypes", json=desired).json()
        return {"scope": self.scope, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if [self.scope, self.code] != [old_state.scope, old_state.code]:
            raise RuntimeError("Cannot change scope/code on datatype")
        remote = self.read(client, old_state)
        remote.pop("href")
        remote.pop("id")
        remote.pop("links")
        remote.pop("version")
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"scope", "code"})
        effective = remote | copy.deepcopy(desired)
        if "referenceData" in remote and "referenceData" in effective:
            # sort the fields since the api changes the order they come back
            remote["referenceData"]["fieldDefinitions"].sort(key=lambda field: field["key"])
            effective["referenceData"]["fieldDefinitions"].sort(key=lambda field: field["key"])
            # sort the values since the api changes the order they come back
            remote["referenceData"]["values"].sort(key=lambda field: field["value"])
            effective["referenceData"]["values"].sort(key=lambda field: field["value"])
        if "acceptableValues" in remote and "acceptableValues" in effective:
            # sort the acceptableValues since the api changes the order they come back
            remote["acceptableValues"].sort()
            effective["acceptableValues"].sort()
        if effective == remote:
            return None
        # check for illegal modifications
        readonly_fields = ["typeValueRange", "unitSchema", "valueType"]
        modified = [field for field in readonly_fields if effective[field] != remote[field]]
        if len(modified) > 0:
            raise RuntimeError(f"Cannot change readonly fields {modified} on datatype")
        if effective["referenceData"]["fieldDefinitions"] != remote["referenceData"]["fieldDefinitions"]:
            raise RuntimeError(
                "Cannot change readonly fields referenceData.fieldDefinitions on datatype"
            )
        # update reference data values if required
        if effective["referenceData"]["values"] != remote["referenceData"]["values"]:
            client.request("PUT", f"/api/api/datatypes/{self.scope}/{self.code}/referencedatavalues",
                json=effective["referenceData"]["values"]
            )
        # update core data if required
        effective.pop("referenceData")
        remote.pop("referenceData")
        if effective != remote:
            client.request("PUT", f"/api/api/datatypes/{self.scope}/{self.code}", json=desired)
        return {"scope": self.scope, "code": self.code}

    @staticmethod
    def delete(client, old_state):
        raise RuntimeError("Cannot delete a datatype definition")

    def deps(self):
        return []
